# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'nvview.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(957, 598)
        self.verticalLayout = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.horizontalLayout.setContentsMargins(-1, -1, 0, -1)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.module_to_nv_button = QtWidgets.QPushButton(Form)
        self.module_to_nv_button.setObjectName("module_to_nv_button")
        self.horizontalLayout.addWidget(self.module_to_nv_button)
        self.write_to_module_button = QtWidgets.QPushButton(Form)
        self.write_to_module_button.setObjectName("write_to_module_button")
        self.horizontalLayout.addWidget(self.write_to_module_button)
        self.write_to_file_button = QtWidgets.QPushButton(Form)
        self.write_to_file_button.setObjectName("write_to_file_button")
        self.horizontalLayout.addWidget(self.write_to_file_button)
        self.write_to_value_set_file_button = QtWidgets.QPushButton(Form)
        self.write_to_value_set_file_button.setObjectName("write_to_value_set_file_button")
        self.horizontalLayout.addWidget(self.write_to_value_set_file_button)
        self.write_to_auto_parameters_button = QtWidgets.QPushButton(Form)
        self.write_to_auto_parameters_button.setObjectName("write_to_auto_parameters_button")
        self.horizontalLayout.addWidget(self.write_to_auto_parameters_button)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.read_from_value_set_file_button = QtWidgets.QPushButton(Form)
        self.read_from_value_set_file_button.setObjectName("read_from_value_set_file_button")
        self.horizontalLayout.addWidget(self.read_from_value_set_file_button)
        self.read_from_file_button = QtWidgets.QPushButton(Form)
        self.read_from_file_button.setObjectName("read_from_file_button")
        self.horizontalLayout.addWidget(self.read_from_file_button)
        self.read_from_module_button = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.read_from_module_button.sizePolicy().hasHeightForWidth())
        self.read_from_module_button.setSizePolicy(sizePolicy)
        self.read_from_module_button.setObjectName("read_from_module_button")
        self.horizontalLayout.addWidget(self.read_from_module_button)
        self.verticalLayout_2.addLayout(self.horizontalLayout)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem1)
        self.label = QtWidgets.QLabel(Form)
        self.label.setObjectName("label")
        self.horizontalLayout_2.addWidget(self.label)
        self.access_level_password = QtWidgets.QLineEdit(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.access_level_password.sizePolicy().hasHeightForWidth())
        self.access_level_password.setSizePolicy(sizePolicy)
        self.access_level_password.setEchoMode(QtWidgets.QLineEdit.Password)
        self.access_level_password.setObjectName("access_level_password")
        self.horizontalLayout_2.addWidget(self.access_level_password)
        self.access_level = QtWidgets.QComboBox(Form)
        self.access_level.setObjectName("access_level")
        self.horizontalLayout_2.addWidget(self.access_level)
        self.set_access_level = QtWidgets.QPushButton(Form)
        self.set_access_level.setObjectName("set_access_level")
        self.horizontalLayout_2.addWidget(self.set_access_level)
        self.current_access_level = Enum(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.current_access_level.sizePolicy().hasHeightForWidth())
        self.current_access_level.setSizePolicy(sizePolicy)
        self.current_access_level.setProperty("label_visible", False)
        self.current_access_level.setObjectName("current_access_level")
        self.horizontalLayout_2.addWidget(self.current_access_level)
        self.line = QtWidgets.QFrame(Form)
        self.line.setFrameShape(QtWidgets.QFrame.VLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.horizontalLayout_2.addWidget(self.line)
        self.auto_read = QtWidgets.QCheckBox(Form)
        self.auto_read.setChecked(True)
        self.auto_read.setObjectName("auto_read")
        self.horizontalLayout_2.addWidget(self.auto_read)
        self.enforce_range_limits_check_box = QtWidgets.QCheckBox(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.enforce_range_limits_check_box.sizePolicy().hasHeightForWidth())
        self.enforce_range_limits_check_box.setSizePolicy(sizePolicy)
        self.enforce_range_limits_check_box.setChecked(True)
        self.enforce_range_limits_check_box.setObjectName("enforce_range_limits_check_box")
        self.horizontalLayout_2.addWidget(self.enforce_range_limits_check_box)
        self.verticalLayout_2.addLayout(self.horizontalLayout_2)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.status_label = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.status_label.sizePolicy().hasHeightForWidth())
        self.status_label.setSizePolicy(sizePolicy)
        self.status_label.setText("")
        self.status_label.setObjectName("status_label")
        self.horizontalLayout_3.addWidget(self.status_label)
        self.label_2 = QtWidgets.QLabel(Form)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout_3.addWidget(self.label_2)
        self.diff_reference_column = QtWidgets.QComboBox(Form)
        self.diff_reference_column.setObjectName("diff_reference_column")
        self.horizontalLayout_3.addWidget(self.diff_reference_column)
        self.verticalLayout_2.addLayout(self.horizontalLayout_3)
        self.searchbox = SearchBox(Form)
        self.searchbox.setObjectName("searchbox")
        self.verticalLayout_2.addWidget(self.searchbox)
        self.tree_view = TreeView(Form)
        self.tree_view.setAlternatingRowColors(True)
        self.tree_view.setObjectName("tree_view")
        self.verticalLayout_2.addWidget(self.tree_view)
        self.verticalLayout.addLayout(self.verticalLayout_2)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.module_to_nv_button.setText(_translate("Form", "Inverter To NV"))
        self.write_to_module_button.setText(_translate("Form", "To Inverter"))
        self.write_to_file_button.setText(_translate("Form", "To File"))
        self.write_to_value_set_file_button.setText(_translate("Form", "To Value Set File"))
        self.write_to_auto_parameters_button.setText(_translate("Form", "To Auto Parameters"))
        self.read_from_value_set_file_button.setText(_translate("Form", "From Value Set File"))
        self.read_from_file_button.setText(_translate("Form", "From File"))
        self.read_from_module_button.setText(_translate("Form", "From Inverter"))
        self.label.setText(_translate("Form", "Access Level:"))
        self.set_access_level.setText(_translate("Form", "Set"))
        self.auto_read.setText(_translate("Form", "Auto Read"))
        self.enforce_range_limits_check_box.setText(_translate("Form", "Enforce Limits Locally"))
        self.label_2.setText(_translate("Form", "Diff Reference Column:"))
from epyqlib.searchbox import SearchBox
from epyqlib.treeview import TreeView
from epyqlib.widgets.enum import Enum
