# This file has been generated


build_system = 'AppVeyor'
build_id = '28463587'
build_number = '385'
build_version = '1.0.385'
job_id = 'lqve17y8c2nlvem5'
job_url = 'https://ci.appveyor.com/project/KyleAltendorf/stlib/build/job/lqve17y8c2nlvem5'
